--NOTES TO SELF

--1.37
--forward loop now enabled by default
--[L] button to load a reference track from O.S. browser
--Textfield added to show name more clearly

--1.38
----stop and restart the transport with L button to `jog` new reference sampl into playing
--fixed bug where [L] button would not initialise correct sample as reference



--note: assert(v [, message])
--Issues an error when the value of its argument v is false
--(i.e., nil or false); otherwise, returns all its arguments.
-- message is an error message; when absent, it defaults to "assertion failed!


-- Dynamically create view hierarchies.
--view:add_child(View child_view)
--view:remove_child(View child_view)
---------------------------------------
---------------------------------------

-----------BEGIN:----------------------

--keybinds

require "functions"

renoise.tool():add_keybinding {
  name = "Global:Tools:`TRT` Toggle Reference Track", 
  invoke = function()toggle_ref_track() end   
}

renoise.tool():add_keybinding {
  name = "Global:Tools:`TRT` Toggle Reference Track With Selected Track", 
  invoke = function()toggle_ref_track_and_selected_track() end   
}

renoise.tool():add_keybinding {
  name = "Global:Tools:`TRT` Go To Reference Track", 
  invoke = function()go_to_ref_track() end   
}

renoise.tool():add_keybinding {
  name = "Global:Tools:`TRT` Adjust Song To Selected Sample Length", 
  invoke = function()extend_song_to_sample_length() end   
}

renoise.tool():add_keybinding {
  name = "Global:Tools:`TRT` Toggle Autoseek On Selected Sample", 
  invoke = function()enable_autoseek() end   
}

renoise.tool():add_keybinding {
  name = "Global:Tools:`TRT` Create Reference Track", 
  invoke = function()
             local song = renoise.song()
             local master_idx = (song.sequencer_track_count+1)
             --where is the master is routed?
             local mst_route = renoise.song().tracks[master_idx].output_routing
             --call function
             create_reference_track(mst_route) 
           end   
}


--Toggle Mono Mix
renoise.tool():add_keybinding {
  name = "Global:Tools:`TRT` Toggle Mix Output Stereo To Mono", 
  invoke = function()mono_tog() end  
}
------------------------------------------------------------------

renoise.tool():add_keybinding {
  name = "Global:Tools:`TRT` Reference Track GUI", 
  invoke = function()main() end   
}

--------------------------------------------------------------------------------
--helper function : custom status message --prefixes tool name and adds brackets 
--------------------------------------------------------------------------------
local function status(message)
  renoise.app():show_status(TOOL_NAME.." Tool: ("..message..")")
end


------------------------------------------
--Set up Preferences file
------------------------------------------
--create xml
local options = renoise.Document.create {
  mid_filter_enabled = true,
}
 --assign options-object to .preferences so renoise knows to load and update it with the tool
renoise.tool().preferences = options
------------------------------------------
------------------------------------------
--variable syntax 
--options.value_a.value

--global
TOOL_NAME = "Toggle Ref. Track"
bypass_timer = false
--UNUSED: default_reference_folder = nil
g_scrub_slider_position = nil

--local global variables for gui
local my_dialog = nil 
local vb = nil
--global timer flag
local timer_running = false
--global load flag
local new_reference_loading = false

--colours
--e.g. For changing vb.views["button_x"].color when states change
COLOR_GREY = {40,40,40}
COLOR_ORANGE ={0xFF,0x66,0x00}
COLOR_YELLOW = {0xE0,0xE0,0x00}
COLOR_BLUE = {0x50,0x40,0xE0}  
COLOR_RED = {0xEE,0x10,0x10}
COLOR_GREEN = {0x20,0x99,0x20}

--to test `vb.views["button_x"].color` we need to check the first entry of the color table
--i.e. if vb.views["button_x"].color[1] == COLOR_GREY_FLAG
COLOR_GREY_FLAG = 40
COLOR_ORANGE_FLAG = 0xFF
COLOR_YELLOW_FLAG = 0xE0
COLOR_BLUE_FLAG = 0x50 
COLOR_RED_FLAG = 0xEE
COLOR_GREEN_FLAG = 0x20


--default Reference Filter presets used by [Low][Mid][High] focus buttons
local low_filter_preset = 
[[
<?xml version="1.0" encoding="UTF-8"?>
<FilterDevicePreset doc_version="11">
  <DeviceSlot type="DigitalFilterDevice">
    <IsMaximized>true</IsMaximized>
    <OversamplingFactor>2x</OversamplingFactor>
    <Model>Chebyshev 8n</Model>
    <Type>
      <Value>0.0</Value>
    </Type>
    <Cutoff>
      <Value>0.125</Value>
    </Cutoff>
    <Q>
      <Value>0.0</Value>
    </Q>
    <Ripple>
      <Value>0.0</Value>
    </Ripple>
    <Inertia>
      <Value>0.0078125</Value>
    </Inertia>
    <ShowResponseView>true</ShowResponseView>
    <ResponseViewMaxGain>18</ResponseViewMaxGain>
  </DeviceSlot>
</FilterDevicePreset>]]

local mid_filter_preset = 
[[
<?xml version="1.0" encoding="UTF-8"?>
<FilterDevicePreset doc_version="11">
  <DeviceSlot type="DigitalFilterDevice">
    <IsMaximized>true</IsMaximized>
    <OversamplingFactor>2x</OversamplingFactor>
    <Model>Chebyshev 8n</Model>
    <Type>
      <Value>1.0</Value>
    </Type>
    <Cutoff>
      <Value>0.35</Value>
    </Cutoff>
    <Q>
      <Value>0.227999985</Value>
    </Q>
    <Ripple>
      <Value>0.0</Value>
    </Ripple>
    <Inertia>
      <Value>0.0078125</Value>
    </Inertia>
    <ShowResponseView>true</ShowResponseView>
    <ResponseViewMaxGain>18</ResponseViewMaxGain>
  </DeviceSlot>
</FilterDevicePreset>]]


local high_filter_preset = 
[[
<?xml version="1.0" encoding="UTF-8"?>
<FilterDevicePreset doc_version="11">
  <DeviceSlot type="DigitalFilterDevice">
    <IsMaximized>true</IsMaximized>
    <OversamplingFactor>2x</OversamplingFactor>
    <Model>Chebyshev 8n</Model>
    <Type>
      <Value>3</Value>
    </Type>
    <Cutoff>
      <Value>0.50</Value>
    </Cutoff>
    <Q>
      <Value>0.227999985</Value>
    </Q>
    <Ripple>
      <Value>0.0</Value>
    </Ripple>
    <Inertia>
      <Value>0.0078125</Value>
    </Inertia>
    <ShowResponseView>true</ShowResponseView>
    <ResponseViewMaxGain>18</ResponseViewMaxGain>
  </DeviceSlot>
</FilterDevicePreset>]]

--local globals to hold cutoff frequencies of the filter (focus) buttons
local low_filter_cutoff_all_enabled_num = 0.125
local mid_filter_cutoff_num = nil --needs to be updated in too --rounding errors
local high_filter_cutoff_all_enabled_num = 0.5
local low_filter_cutoff_mid_disabled_num = 0.32
local high_filter_cutoff_mid_disabled_num = 0.33


--gui function (mid band enable checkbox): resizes buttons, hides/ unhides [MID] button
--used in vb.views["mid checkbox"] notifier and on tool first initilaisation: in main()
--------------------------------------------------------------------
local function modify_filter_cutoffs_low_and_high_buttons(val)--bool
--------------------------------------------------------------------
  if val == false then --only two filters visible
    vb.views["mid button"].active = false
    vb.views["mid button"].visible = false
    --resize low and high buttons to fill gui row
    vb.views["low button"].width = 81
    vb.views["high button"].width = 81
  else
    vb.views["low button"].width = 54
    vb.views["high button"].width = 54
    --show and activate mid button
    vb.views["mid button"].active = true
    vb.views["mid button"].visible = true
  end
end


--main gui function
------------------------------------
------------------------------------
function main()
------------------------------------
------------------------------------
  --toggle dialog so if its already open then close it.  The timer will catch
  --that it is closed and do the housekeeping like removing notifiers (and timer itself)
  if (my_dialog and my_dialog.visible) then 
    my_dialog:close()
    return
  end
  
  --renoise song object
  local song = renoise.song()

  --set globals
  -------------
  --assign global vb to viewbuilder
  vb = renoise.ViewBuilder()
  
  --get position for scrub slider if ref track present
  ----------------------------------------------------
  local ref_inst_index = get_reference_track_instrument_index()
  local ref_sample_length = 1
  if ref_inst_index ~= nil then
    
    local ref_inst = song:instrument(ref_inst_index)
    --get sample
    if #ref_inst.samples > 0 then
      local ref_sample = ref_inst:sample(1)
      --get slice marker pos
      if #ref_sample.slice_markers > 0 then
        --update global to slice marker 1
        g_scrub_slider_position = ref_sample.slice_markers[1] --no method available, only table
        --get sample length to set scrub slider max value to
        ref_sample_length = ref_sample.sample_buffer.number_of_frames
      end
    end
  else
    --default to 0
    g_scrub_slider_position = 1
  end
    
  -------------------------------------------------------------------------------------------
  --GUI
  -------------------------------------------------------------------------------------------  
  local dialog_content = 
    vb:vertical_aligner {
      margin = 10,
      
     vb:horizontal_aligner{
       mode = "center",
       margin = 4,
       
      --ref wav name 
      -------------- 
      vb:textfield{
        width = 120,
        text = "",
        id = "wav name",
        notifier = function(text)
           
           if text == "- NO REF SAMPLE -" then
             return
           end 
           
           if text == "" then
             status("Sample Requires A Name")
             return
           end 
           
           --check for ref track
           local ref_track_idx = get_ref_track_index_via_ref_gainer()
           
           if ref_track_idx == nil then
             vb.views["wav name"].text = "- NO REF SAMPLE -"
             return
           end
           
           local ref_note_inst = nil
           --get ref .wav instrument slot index
           local ref_track_note_first_col = song:pattern(1):track(ref_track_idx):line(1):note_column(1)
           if ref_track_note_first_col.instrument_value ~= 255 then--255 is empty
            --get instrument name
            ref_note_inst = song:instrument(ref_track_note_first_col.instrument_value + 1)
           end
           
           ref_note_inst.name = text
           
        
        
          end
      },
       
      vb:button{
       text = "L",
       notifier = function()
         local app = renoise.app()
         local song = renoise.song()
         
         --set global flag
         new_reference_loading = true
         
         local ref_properties = table_of_ref_track_properties()--.ref_track_index .first_note .ref_instrument
         
         --will be instrument object      
         local ref_note_inst = nil
         --will be instrument index
         local ref_note_inst_idx = nil
         --get ref track Index
         local ref_track_idx = ref_properties.ref_track_index
         
         --1) CASE NO REFERENCE TRACK FOUND
         if ref_track_idx == nil then 
           
           --make an instrument at pos 1 and add an empty sample at sample pos 1
           song:insert_instrument_at(1)
           --select instrument slot with the reference file-- create_reference_track(mst_route) works on/uses selected instrument
           song.selected_instrument_index = 1
           ref_note_inst = song:instrument(1)
           ref_note_inst:insert_sample_at(1) --empty  sample will be overwritten
          
           --Create ref track
           --------------------
           local master_idx = (song.sequencer_track_count+1)
           --where is the master is routed?
           local mst_route = renoise.song().tracks[master_idx].output_routing
           --call function
           create_reference_track(mst_route)
           --ref track will now be track 1
           ref_track_idx = 1
           
           --update play btn text
           vb.views["play ref"].text = "REF PLAYING"
           
           --housekeeping so buttons are refreshed back to default if gui is still open when ref track deleted manually in renoise
           if vb.views["mono button"].color[1] == COLOR_RED_FLAG then
             mono_tog() --toggle mono back to default state
             vb.views["mono button"].color = COLOR_GREY
           end
           --reset LOW/MID/HIGH
           vb.views["low button"].color = COLOR_GREY
           vb.views["mid button"].color = COLOR_GREY
           vb.views["high button"].color = COLOR_GREY
      
         --++++++++++++++++++++++++++++++++++++++++++
         --2)CASE REFERENCE TRACK FOUND
         else
          
           --get ref .wav instrument slot index
           local ref_track_note_first_col = ref_properties.first_note --song:pattern(1):track(ref_track_idx):line(1):note_column(1)
           --check to see make sure first note has an instrument number
           if ref_track_note_first_col.instrument_value ~= 255 then--255 is empty
             --get instrument
             ref_note_inst = song:instrument(ref_track_note_first_col.instrument_value + 1)
             ref_note_inst_idx = ref_track_note_first_col.instrument_value + 1
           else --show dialog for user to choose coreect sample slot
             ref_note_inst = song.selected_instrument --i.e. if no inst no. in column
             
             local instruments_tab = {}
            -- table.insert(instruments_tab,"Cancel")
             for i = 1,#song.instruments do
               instruments_tab[i+1] = tostring(i)..": "..song:instrument(i).name
             end
             
             
             local prompt_text = "Reference Track Is Missing Instrument Data\nChoose The Correct Instrument In The Popup List Below That Has\nThe Current Reference Sample.\n\nNOTE: The New Sample Chosen In the Next Dialog Will Overwrite It.\n(As Expected After Pressing [L] Button)\n\nAlternatively Press Cancel To Quit\n"
             local vb = renoise.ViewBuilder()
             local prompt_content = vb:column{
                    margin = 6,
                     vb:text{
                      text = prompt_text
                      },
                     vb:popup{
                      width = 350,
                      items = instruments_tab,
                      id = "new inst tab",
                     },
                     vb:text{
                      text = "", --spacer
                     }
                    }
                    
                    

             local ref_note_string = renoise.app():show_custom_prompt("Choose Instrument - Reference Track Is Missing Data ",prompt_content , {"Continue","Cancel"})
             
             if ref_note_string == "Cancel" then --user cancelled with cancel button
               return
             end  
             
             ref_note_inst_idx = vb.views["new inst tab"].value + 1 --adjust +1 to renoise inst idx
             
              --update  ref instrument object and index
             ref_note_inst_idx = ref_note_inst_idx - 1 --+1 due to cancel button
             ref_note_inst =  song:instrument(ref_note_inst_idx)
           end
           
           --update ref note in renoise: make sure note is set to C#4 so slice marker can be used (older versions of tool added c-4 which may still be present)         
           --updated below
           --ref_track_note_first_col.note_value = 49
        
         
         end
         
         --add a sample slot and sample data if none are present
         if #ref_note_inst.samples == 0 then
           ref_note_inst:insert_sample_at(1)
           ref_note_inst:sample(1).sample_buffer:create_sample_data(44100, 16, 1, 10)  
         end
         
         
        --clear all slice markers from current sample then re-add C#4 (gets deleted when slice marker deleted)
        --and re-add instrument number
        
        --loop to clear old slice markers
        local ref_sample = nil
        if ref_note_inst ~= nil then
           ref_sample = ref_note_inst:sample(1)
           --clear xs slice markers if sample data found
           if ref_sample.sample_buffer.has_sample_data == true then
  
             --RAN INTO SLICE MARKER BUG shouldn`t effect tool******Reported Nov 2017 on forum
             
             --store first note
             local ref_track_note_first_col = song:pattern(1):track(ref_track_idx):line(1):note_column(1)
             
             --loop to delete any extra slice markers --NOTE: THIS ALSO DELETES REFERENCE NOTE
             for i = #ref_sample.slice_markers, 1,-1 do
               ref_sample:delete_slice_marker(ref_sample.slice_markers[i])
             end
             
             --re-instate note that got deleted by deleting slice markers
             ref_track_note_first_col.note_value = 49
             ref_track_note_first_col.instrument_value = ref_note_inst_idx - 1
             --re-instate first slice marker at beginning of sample
             ref_sample:insert_slice_marker(1)
           end
         end
         
         ---load sample
         ----------------
         local sample_path = renoise.app():prompt_for_filename_to_read({"*"}, "load") --returns empty string if cancel pressed
         --return if no path found
         if sample_path == "" then
           status("No Sample Was Chosen To load")
           --set global flag
           new_reference_loading = false
           return
         end
         
         --return on failed load + write to status
         --------------------------
         local load_successful = ref_note_inst.samples[1].sample_buffer:load_from(sample_path)
         if load_successful == false then
           status("Renoise Was Unable To Load Sample")
           --set global flag
           new_reference_loading = false
           return
         end 
         
         -------------------------------------------------
         -- If we get here, a new sample has been loaded--
         -------------------------------------------------
         
         --reset scrub slider
         vb.views["scrub slider"].value = 1
         --reset scrub_slider global for timer
         g_scrub_slider_position = 1
         --update range/ max value of to fit new sample
         vb.views["scrub slider"].max = ref_sample.sample_buffer.number_of_frames
         
         --add one slice marker at sample 1
         if #ref_sample.slice_markers < 1 then
           ref_sample:insert_slice_marker(1)
         end 
         
         --make sure note for slice marker is added in pattern 
         local ref_track_note_first_col = song:pattern(1):track(ref_track_idx):line(1):note_column(1)
         --set first note to C#4
         ref_track_note_first_col.note_value = 49
         --if instrument number is empty then set it to the selected instrument in renoise
         if ref_note_inst_idx == nil then
           ref_track_note_first_col.instrument_value = (song.selected_instrument_index - 1)
         else
           ref_track_note_first_col.instrument_value = (ref_note_inst_idx - 1)
         end        
       
         --make sure autoseek is turned on
         ---------------------------------
         ref_note_inst:sample(1).autoseek = true
         
         --switch on forward loop mode for sample
         ---------------------------------------- 
         ref_note_inst:sample(1).loop_mode = renoise.Sample.LOOP_MODE_FORWARD
          
         --get sample name from path
         ---------------------------
         local next_back_slash = nil
         --default name is full sample  path
         local sample_name = sample_path
         
         --loop to find "\\" single backslash and remove all text before and including the
         --last backslash, leaving the sample name (/ forward slashes for unix systems)
         for i = 1,#sample_name do
           next_back_slash = sample_name:find("\\") or sample_name:find("/") --/for mac and Linux
           --back_slash found so truncate section
           if (next_back_slash ~= nil) then
             sample_name = sample_name:sub(next_back_slash+1,#sample_name)
             --reset flag
             next_back_slash = nil
           end
         end
         
         --set instrument name
         ----------------------
         ref_note_inst.name = sample_name
         --stop and restart the transport to `initialise` the sample new to play ( deals with quirk of renoise autoseek behaviour)
         if song.transport.playing == true then
           --toggle the transport
           song.transport:stop()
           song.transport:start_at(song.selected_line_index)
         end
         --set global flag
          new_reference_loading = false
         
       end
      },
      
      
     },
     
     
      vb:horizontal_aligner{
       mode = "center",
       margin = 4,
        
        vb:minislider{
          width = 140,
          id = "scrub slider",
          min = 1,
          max = ref_sample_length,
          value = g_scrub_slider_position,
          notifier = function(value)
              
              if timer_running == true then
                return
              end
              ---check global to see if sample being loaded 
              --(loading ref sample deletes the reference note and resets this slider which can cause errors in this function)
              if new_reference_loading == true then
                return
              end
              --song
              local song = renoise.song()
              
              --get ref instrument index / return if no instrument index is found in pattern note
              local ref_inst_idx = get_reference_track_instrument_index() --fn fires warning to user when nil is returned
              if ref_inst_idx == nil then
                --assert(nil,"ref inst index / track not present")
                 vb.views["scrub slider"].value = 1
                 g_scrub_slider_position = 1
                return
              end
              
              --get reference instrument object
              local ref_instrument = song:instrument(ref_inst_idx)
              
              --check if sample exists in reference instrument and show error if not and return
              if #ref_instrument.samples == 0 then
                ref_inst_idx = ref_inst_idx - 1
                --convert index to string
                local index_string = tostring(ref_inst_idx)
                index_string = string.format("%X", index_string)
                --add leading zero if needed
                if #index_string < 2 then
                  index_string = "0"..index_string
                end
                renoise.app():show_error("No samples are present in the reference instrument,\nInstrument Number: "..index_string)
                return
              end
              --get sample
              local ref_sample = ref_instrument:sample(1)
              --get ref note so we can re-add the norte below (it gets deleted when modifying the slice slider value)
              local ref_track_note_first_col = get_reference_track_first_pattern_note()
              --check for sample data
              if ref_sample.sample_buffer.has_sample_data == true then
               --check for slice markers
               if #ref_sample.slice_markers > 0 then
                 --update range/ max value of `this` minslider
                 vb.views["scrub slider"].max = ref_sample.sample_buffer.number_of_frames
                 --catch when short sample may have been loaded over longer one--lag in update of the max value 
                 if value < ref_sample.sample_buffer.number_of_frames then
                   ref_sample:move_slice_marker(ref_sample.slice_markers[1], value)
                   g_scrub_slider_position = value
                   --make sure uatoseek is turned on
                   ref_sample.autoseek = true
                 else
                   --no need for warning as next move of slider will work
                   return
                 end
                 --re-add note into renoise
                 ref_track_note_first_col.note_value = 49 --C#4
                 ref_track_note_first_col.instrument_value = ref_inst_idx - 1
                 --toggle the transport to restart (jog) sample playing
                 if song.transport.playing == true then
                   song.transport:stop()
                   song.transport:start_at(song.selected_line_index)
                 end
               end
             end
           end
          },
       },
      --ROW 1
      ---------
      vb:horizontal_aligner{
        mode = "center",
        
        vb:button { 
           width = 140,
           height = 30,
           color = COLOR_GREY,
           text = "Create Reference Track:",
           id = "create ref btn",
           notifier = function()
                          
                         local song = renoise.song()  
                         
                         --1) Dual function button changes to "Go To Reference Track" after one is created
                         ---------------------------------------------------------------------------------
                         if vb.views["create ref btn"].text == "Go To Reference Track" then
                           song.selected_track_index = get_ref_track_index_via_ref_gainer()
                           return
                         end
                         --2)Create ref track
                         --------------------
                         local master_idx = (song.sequencer_track_count+1)
                        --where is the master is routed?
                         local mst_route = renoise.song().tracks[master_idx].output_routing
                         --call function
                         create_reference_track(mst_route)
                         --update play btn text
                         vb.views["play ref"].text = "REF PLAYING"
                         
                         --housekeeping so buttons are refreshed back to default if gui is still open when ref track deleted manually in renoise
                         if vb.views["mono button"].color[1] == COLOR_RED_FLAG then
                           mono_tog() --toggle mono back to default state
                           vb.views["mono button"].color = COLOR_GREY
                         end
                         --reset LOW/MID/HIGH
                         vb.views["low button"].color = COLOR_GREY
                         vb.views["mid button"].color = COLOR_GREY
                         vb.views["high button"].color = COLOR_GREY
                         
                        

                       end
            },--button 1
           },--row (horizontal_aligner{) 1
            
            --text line shows instruction on initial screen
            -------
            vb:text{
               text = "",
               id = "start text", 
             },
             
         --ROW 2
         --------
         --controls column --this whole columns contains lower part of GUI and gets hidden until erf track is present
         vb:column{
           id = "controls", 
            
           vb:row{
             vb:text{
               text = "Toggle:                   Single Trk:"--text info row
             },
             vb:checkbox{
              id =  "toggle current box",
              notifier = function(value)
                           local song = renoise.song()
                           if (vb.views["play ref"].color[1] == COLOR_GREY_FLAG) then
                             local selected_track = renoise.song().tracks[song.selected_track_index] 
                             
                           --[[
                             --value false so solo track enabled
                             --sometimes a track will return soloed when really it`s not (API bug? has been reported)
                             --so we need to loop and flag other unmuted sequencer tracks
                             local not_solo_flag = false
                             --loop tracks    
                             for track = 1,#song.tracks do
                               --skip selected track
                               if track ~= song.selected_track_index then
                                 --make sure sequencer track
                                 if song:track(track).type == renoise.Track.TRACK_TYPE_SEQUENCER then
                                   --see if unmuted
                                   if song:track(track).mute_state == renoise.Track.MUTE_STATE_ACTIVE then
                                     not_solo_flag = true
                                   --  print("true")
                                     break
                                   end
                                 end   
                               end
                             end--]]
                             
                             --value true so solo track enabled
                             --sometimes a track will return soloed when really it`s not
                             -- (API bug? has been reported)
                             
                             --if checkbox ticked then solo current track
                             if (value == true) then 
                                selected_track:mute() --hack to make sure solos correctly
                                selected_track:solo()
                               return
                             end
                             
                             --if checkbox un-ticked then unmute all tracks
                             if (value == false) then
                              for track = 1,#song.tracks do
                                if song:track(track).type ~= renoise.Track.TRACK_TYPE_MASTER then
                                  song:track(track).mute_state = renoise.Track.MUTE_STATE_ACTIVE
                                end
                               end
                             end
                           end
                         end
              },
             }, --ROW 2
           
           --ROW 3----
           ----------- 
           vb:row{
             vb:button { 
              width = 160,
              height = 48,
              color = COLOR_GREY,
              text = "SONG PLAYING",
              id = "play ref",
              notifier = function()
                           --toggle the reference track
                           if vb.views["toggle current box"].value == false then
                             toggle_ref_track()
                            
                           else
                             toggle_ref_track_and_selected_track()
                           end
                           
                            --run timer to prevent gui lag
                            timer()
                           
                           --check if active or not
                           local song = renoise.song()
                           local ref_track_idx = get_ref_track_index_via_ref_gainer()
                           --return if no ref track found
                           if ref_track_idx == nil then
                             return
                           end
                           --get ref track object
                           local ref_track = song:track(ref_track_idx)
                          
                          for devs = 2, #ref_track.devices do
                            --find toggle gainer
                            if ref_track.devices[devs].display_name == "Reference Gainer" then 
                              --1): REF TRACK PLAYING
                              if ref_track.devices[devs].is_active ~= true then
                                vb.views["play ref"].color = COLOR_GREEN
                                vb.views["play ref"].text = "REF PLAYING"
                                return
                              else  --2):  REF TRACK MUTED
                                vb.views["play ref"].color = COLOR_GREY
                                vb.views["play ref"].text = "SONG PLAYING"
                                return
                              end
                            end
                          end 
                        end
              },
             },--row 3
             
            vb:row{
             --text line above filter buttons
             vb:text{
               text = "Focus:                             Mid:"
              },
              
              vb:checkbox{
                value = options.mid_filter_enabled.value,
                id = "mid checkbox",
                notifier = function(value)
                    
                   
                    --run
                    modify_filter_cutoffs_low_and_high_buttons(value)
                    --update tool preferences
                    options.mid_filter_enabled.value = value
                    
                    --run timer to update gui without lag
                    timer()

                  
                  end,
                
              
              
              }
              
              
              
              
             },
            --ROW 4---- Filter buttons LOW/MID/HIGH
            -----------
            vb:row{
             vb:button{
                width = 54,
                height = 30,
                text = "LOW",
                id = "low button",
                color = COLOR_GREY,
                notifier = function()
                            local song = renoise.song()
                            local master_index = #song.tracks - song.send_track_count
                            local master_track = song:track(master_index)
                            
                            --find reference track and master track
                            local ref_track_index = false
                            for track = 1, #song.tracks do
                              for devs = 2, #song:track(track).devices do
                                --find toggle gainer
                                if song:track(track).devices[devs].display_name == "Reference Gainer" then 
                                  ref_track_index = track
                                  break
                                end
                              end
                            end
                            --return if no reference track is found
                            if ref_track_index == false then
                              status("No Reference track found")
                              return
                            end
                            
                            local ref_track = song:track(ref_track_index)
                            local ref_filter = nil
                            --1) loop backwards through ref track devices to find filter
                            for devs = 2, #ref_track.devices do
                              --find toggle gainer
                              if ref_track.devices[devs].display_name == "Reference Filter" then 
                                ref_filter = ref_track.devices[devs]
                                break
                              end
                            end
                            
                            local mas_ref_filter = nil
                            --2) loop backwards through master track devices to find filter
                            for devs = 2, #master_track.devices do
                              --find toggle gainer
                              if master_track.devices[devs].display_name == "Reference Filter" then 
                                mas_ref_filter = master_track.devices[devs]
                                break
                              end
                            end
                            
                            --add filter if not present
                             --add filter at end of ref track if not present
                            if ref_filter == nil then
                              ref_filter = ref_track:insert_device_at("Audio/Effects/Native/Digital Filter", #ref_track.devices + 1 )
                              ref_filter.display_name = "Reference Filter"
                              ref_filter.is_active = false --will toggle back on next
                            end
                            
                            --add master filter if not present
                             --add filter at end of ref track if not present
                            if mas_ref_filter == nil then
                              mas_ref_filter = master_track:insert_device_at("Audio/Effects/Native/Digital Filter", #master_track.devices + 1)
                              mas_ref_filter.display_name = "Reference Filter"
                              mas_ref_filter.is_active = false --will toggle back on next
                            end
                            
                           --toggle state
                           if vb.views["low button"].color[1] == COLOR_GREY_FLAG then
                              vb.views["low button"].color = COLOR_ORANGE
                              vb.views["mid button"].color = COLOR_GREY
                              vb.views["high button"].color = COLOR_GREY
                              ref_filter.is_active = true
                              mas_ref_filter.is_active = true
                           else
                             vb.views["low button"].color = COLOR_GREY
                             vb.views["mid button"].color = COLOR_GREY
                              vb.views["high button"].color = COLOR_GREY
                             ref_filter.is_active = false
                             mas_ref_filter.is_active = false
                           end
                            
                            --set to low filter preset
                            ref_filter.active_preset_data = low_filter_preset
                            mas_ref_filter.active_preset_data = low_filter_preset
                            
                             --get cutoff freq if mid button disabled or set to nil if not
                            if vb.views["mid checkbox"].value == false then
                              ref_filter:parameter(2).value = low_filter_cutoff_mid_disabled_num
                              low_filter_cutoff_mid_disabled_num = ref_filter:parameter(2).value
                            else
                              ref_filter:parameter(2).value = low_filter_cutoff_all_enabled_num
                            end

                           end
                },
              
              vb:button{
                width = 54,
                height = 30,
                text = "MID",
                id = "mid button",
                color = COLOR_GREY,
                notifier = function()
                              local song = renoise.song()
                              local master_index = #song.tracks - song.send_track_count
                              local master_track = song:track(master_index)
                              
                              --find reference track
                              local ref_track_index = false
                              for track = 1, #song.tracks  do
                                for devs = 2, #song:track(track).devices do
                                  --find toggle gainer
                                  if song:track(track).devices[devs].display_name == "Reference Gainer" then 
                                    ref_track_index = track
                                    break
                                  end
                                end
                              end
                              --return if no reference track is found
                              if ref_track_index == false then
                                status("No Reference track found")
                                return
                              end
                              
                              --1) find ref filter
                              local ref_track = song:track(ref_track_index)
                              local ref_filter = nil
                              --loop backwards through devices to find filter
                              for devs = 2, #ref_track.devices do
                                --find toggle gainer
                                if ref_track.devices[devs].display_name == "Reference Filter" then 
                                  ref_filter = ref_track.devices[devs]
                                  break
                                end
                              end
                              
                               local mas_ref_filter = nil
                              --2) loop backwards through master track devices to find filter
                              for devs = 2, #master_track.devices do
                                --find toggle gainer
                                if master_track.devices[devs].display_name == "Reference Filter" then 
                                  mas_ref_filter = master_track.devices[devs]
                                  break
                                end
                              end
                              
                               --add filter at end of ref track if not present
                              if ref_filter == nil then
                                ref_filter = ref_track:insert_device_at("Audio/Effects/Native/Digital Filter", #ref_track.devices+1)
                                ref_filter.display_name = "Reference Filter"
                                ref_filter.is_active = false --will toggle back on next
                              end
                              
                              --add master filter if not present
                             --add filter at end of ref track if not present
                            if mas_ref_filter == nil then
                              mas_ref_filter = master_track:insert_device_at("Audio/Effects/Native/Digital Filter", #master_track.devices+1)
                              mas_ref_filter.display_name = "Reference Filter"
                              mas_ref_filter.is_active = false --will toggle back on next
                            end
                              
                             --toggle state
                             if vb.views["mid button"].color[1] == COLOR_GREY_FLAG then
                                vb.views["mid button"].color = COLOR_ORANGE
                                vb.views["low button"].color = COLOR_GREY
                                vb.views["high button"].color = COLOR_GREY
                                ref_filter.is_active = true
                                mas_ref_filter.is_active = true
                             else
                               vb.views["mid button"].color = COLOR_GREY
                               vb.views["low button"].color = COLOR_GREY
                                vb.views["high button"].color = COLOR_GREY
                               ref_filter.is_active = false
                               mas_ref_filter.is_active = false
                             end
                              
                              --set to mid filter preset
                              ref_filter.active_preset_data = mid_filter_preset
                              mas_ref_filter.active_preset_data = mid_filter_preset
                              --update global --rounding bug??
                              mid_filter_cutoff_num = ref_filter:parameter(2).value
                             
                            
                           end
                },
               
               vb:button{
                width = 54,
                height = 30,
                text = "HIGH",
                id = "high button",
                color = COLOR_GREY,
                notifier = function()
                             local song = renoise.song()
                             local master_index = #song.tracks - song.send_track_count
                             local master_track = song:track(master_index)
                            
                            --find reference track
                            local ref_track_index = false
                            for track = 1, #song.tracks  do
                              for devs = 2, #song:track(track).devices do
                                --find toggle gainer
                                if song:track(track).devices[devs].display_name == "Reference Gainer" then 
                                  ref_track_index = track
                                  break
                                end
                              end
                            end
                            --return if no reference track is found
                            if ref_track_index == false then
                              status("No Reference track found")
                              return
                            end
                            
                            --1)
                            local ref_track = song:track(ref_track_index)
                            local ref_filter = nil
                            --loop backwards through devices to find filter
                            for devs = 2, #ref_track.devices do
                              --find toggle gainer
                              if ref_track.devices[devs].display_name == "Reference Filter" then 
                                ref_filter = ref_track.devices[devs]
                                break
                              end
                            end
                            
                            local mas_ref_filter = nil
                            --2) loop backwards through master track devices to find filter
                            for devs = 2, #master_track.devices do
                              --find toggle gainer
                              if master_track.devices[devs].display_name == "Reference Filter" then 
                                mas_ref_filter = master_track.devices[devs]
                                break
                              end
                            end
                            
                             --add filter if not present
                             --add filter at end of ref track if not present
                            if ref_filter == nil then
                              ref_filter = ref_track:insert_device_at("Audio/Effects/Native/Digital Filter", #ref_track.devices+1)
                              ref_filter.display_name = "Reference Filter"
                              ref_filter.is_active = false --will toggle back on next
                            end
                            
                             --add master filter if not present
                             --add filter at end of ref track if not present
                            if mas_ref_filter == nil then
                              mas_ref_filter = master_track:insert_device_at("Audio/Effects/Native/Digital Filter", #master_track.devices+1)
                              mas_ref_filter.display_name = "Reference Filter"
                              mas_ref_filter.is_active = false --will toggle back on next
                            end
                            
                           --toggle state
                           if vb.views["high button"].color[1] == COLOR_GREY_FLAG then
                              vb.views["high button"].color = COLOR_ORANGE
                              vb.views["mid button"].color = COLOR_GREY
                              vb.views["low button"].color = COLOR_GREY
                              ref_filter.is_active = true
                              mas_ref_filter.is_active = true
                           else
                             vb.views["high button"].color = COLOR_GREY
                             vb.views["mid button"].color = COLOR_GREY
                              vb.views["low button"].color = COLOR_GREY
                             ref_filter.is_active = false
                             mas_ref_filter.is_active = false
                           end
                            
                            --set to high filter preset
                            ref_filter.active_preset_data = high_filter_preset
                            mas_ref_filter.active_preset_data = high_filter_preset 
                            
                            --get cutoff freq if mid button disabled or set to nil if not
                            if vb.views["mid checkbox"].value == false then
                              ref_filter:parameter(2).value = high_filter_cutoff_mid_disabled_num
                              high_filter_cutoff_mid_disabled_num = ref_filter:parameter(2).value
                            else
                              ref_filter:parameter(2).value = high_filter_cutoff_all_enabled_num
                            end
                            
                           end
                },
              },--row 4
           --[[   
           vb:row{ --debug
             vb:popup{
               items = available_filter_types,
               width = 120,
               notifier = function(value)
                            substitute_filter_type(value)  
                          end
             }
           },--]]
              
              vb:text{
               text = "Ref Track Volume:"--             Mono:",
              },
              
              
            --ROW 5---- slider and mono button
            ----------- 
            vb:row{
              vb:slider{
                width = 138,
                height = 20,
                min = renoise.song().tracks[1].postfx_volume.value_min,
                max = renoise.song().tracks[1].postfx_volume.value_max,
                value = 1,
                id = "volume slider",
                notifier = function(value)
                             bypass_timer = true
                             local song = renoise.song
                             renoise.song().tracks[1].postfx_volume.value = value
                             bypass_timer = false
                           end
              
              },
              
            -- mono button
          --  vb:row{
              vb:button{
                width = 20,
                height = 20,
                text = "M",
                id = "mono button",
                notifier = function(value)
                             mono_tog()
                             --run timer to update GUi
                             timer()
                           end
               },
            -- },
             },
            },--controls column
           } --Vertical aligner --END of GUI content
  --------------------------------------------------
  --------------------------------------------------
   
   
   --adjust initial gui to relect saved preferences
   if options.mid_filter_enabled.value ==false then
     modify_filter_cutoffs_low_and_high_buttons(false)
   end
   
         
  --------------
  --key Handler
  --------------
  local function my_keyhandler_func(dialog,key)
     --toggle lock focus hack, allows pattern ed to get key input
     renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
     renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
     return key
  end
   
  ------------------------------------------------------------------------------------------
  ------------------------------------------------------------------------------------------
  --Script dialog
  my_dialog = renoise.app():show_custom_dialog(
      TOOL_NAME, dialog_content,my_keyhandler_func)

 
  --add timer to fire once every 80ms
  if not renoise.tool():has_timer(timer) then
    renoise.tool():add_timer(timer,80)
  end
 
  --------------------------------------------    
  --close dialog function ON NEW SONG
  --------------------------------------------
  local function closer(d)
    my_dialog = nil    
    if d and d.visible then
      d:close()
      -- remove_notifiers()
      if renoise.tool():has_timer(timer) then
        renoise.tool():remove_timer(timer)
      end
    end
  end
  -- notifier to close dialog on load new song
  renoise.tool().app_release_document_observable:add_notifier(closer,my_dialog)
  -------------------------------------------------------------------------------
  
  --first run of timer
   timer()
end
-----------------------------------------------
----------end of main()------------------------
-----------------------------------------------

function timer()


  if bypass_timer == true then
    return
  end
  
  --check global flag
  if new_reference_loading == true then
    return
  end
  
  
  --timer flag true 
  --(needs to be false just before evrey subsequent return statement)
  timer_running = true
  
  local song = renoise.song()

  local ref_track_idx = get_ref_track_index_via_ref_gainer()
  
  --1) CASE: NO REFERENCE TRACK FOUND --SET UP FOR FIRTS USE
  ----------------------------------------------------------
  if ref_track_idx == nil then
    vb.views["wav name"].active = false
    --clear wav-name text
    vb.views["wav name"].text = ""
    --show help text
    vb.views["start text"].text =
    "   Load a ref .wav in Inst. box"--\n  or Load via `L` button above"
    --make sure create button is active (Also doubles as `Go To Reference Track` button and can be made inactive in that state, i.e. when ref track is selected)
    vb.views["create ref btn"].active = true
    --set start button to "Create Reference Track"
    vb.views["create ref btn"].text = "Create Reference Track"
    --hide main controls
    vb.views["controls"].visible = false
    --Delete Reference filter on master track
    local master_index = #song.tracks - song.send_track_count
    local master_track = song:track(master_index)
    
    for devs = #master_track.devices,2 ,-1 do
      if master_track.devices[devs].display_name == "Reference Filter" then
        master_track:delete_device_at(devs)
      end
    end
    
    --hide and disable scrub slider
    vb.views["scrub slider"].active = false
    vb.views["scrub slider"].visible = false
    
  --2) CASE: REFERENCE TRACK IS FOUND --ADJUST GUI TO CURRENT STATE
  -----------------------------------------------------------------
  else
    
    local ref_track = song:track(ref_track_idx)
   
    --Disable `Go To Reference Track` button if on reference track/ else enable
    if (song.selected_track_index == ref_track_idx) then
      vb.views["create ref btn"].text = "REF TRACK SELECTED"
     -- vb.views["create ref btn"].color = COLOR_GREEN
      vb.views["create ref btn"].active = false
    else
      vb.views["create ref btn"].text = "Go To Reference Track"
      vb.views["create ref btn"].color = COLOR_GREY
      vb.views["create ref btn"].active = true
    end
    
    ---UPDATE wav name text
    local ref_note_inst = ""
    --get ref .wav instrument slot index
    local ref_track_note_first_col = song:pattern(1):track(ref_track_idx):line(1):note_column(1)
    if ref_track_note_first_col.instrument_value ~= 255 then--255 is empty
      --get instrument name
      ref_note_inst = song:instrument(ref_track_note_first_col.instrument_value + 1).name
    end
    
    vb.views["wav name"].active = true
    vb.views["wav name"].text = ref_note_inst -- string.sub(ref_note_inst,1,21) 
    --remove help text
    vb.views["start text"].text = ""
    --show main controls
    vb.views["controls"].visible = true
    
    --REF IS ACTIVE AND PLAYING
    --set the color for the play button
    if song:track(ref_track_idx).name == "`REF` ACTIVE" then
      vb.views["play ref"].color = COLOR_GREEN
      --check if reference track is selected, disable button if single track toggle mode
      if (song.selected_track_index == ref_track_idx) and (vb.views["toggle current box"].value == true) then
        vb.views["play ref"].active = false
        vb.views["play ref"].text = "Can`t Toggle Ref With Itself"
      else --else play ref button back to normal
        vb.views["play ref"].active = true
        vb.views["play ref"].text = "REF PLAYING"
      end
      --enable and show scrub slider
      vb.views["scrub slider"].active = true
      vb.views["scrub slider"].visible = true
      
      --*********************************************---------------
      --uodate scrub slider position --needs initial values
      
      if (g_scrub_slider_position ~= nil) and (g_scrub_slider_position <= vb.views["scrub slider"].max) then
        vb.views["scrub slider"].value = g_scrub_slider_position
      end
      
      
      

    else --REF IS INACTIVE (SONG PLAYING) 
      
      --check if reference track is selected, disable button if single track toggle mode
      if (song.selected_track_index == ref_track_idx) and (vb.views["toggle current box"].value == true) then
        vb.views["play ref"].active = false
        vb.views["play ref"].text = "Can`t Toggle Ref With Itself"
      else --else play ref button back to normal
        vb.views["play ref"].active = true
        vb.views["play ref"].text = "SONG PLAYING"
      end 
      --disable and hide scrub slider
      vb.views["scrub slider"].active = false
      vb.views["scrub slider"].visible = false
      --toggle button to grey
      vb.views["play ref"].color = COLOR_GREY
    end
    ---------------------------------------------
    
    
    --TODO UPDATE minislider
--[[
    if ref_sample.sample_buffer.has_sample_data == true then

     if #ref_sample.slice_markers > 0 then
       vb.views["scrub slider"].max = ref_sample.sample_buffer.number_of_frames
   --]]  

    
    
    --check for state of filter (LOW,MID,HIGH)
    ------------------------------------------
    
    local ref_filter = nil
    --loop backwards through devices to find filter
    for devs = 2, #ref_track.devices do
      --find toggle gainer
      if ref_track:device(devs).display_name == "Reference Filter" then 
        ref_filter = ref_track:device(devs)
        break
      end
    end
    
     --find Mono Mix  device
    for devs = 2, #ref_track.devices do
      if (ref_track:device(devs).display_name == "Mono Mix") then 
        if (ref_track:device(devs).is_active == true) then
          vb.views["mono button"].color = COLOR_RED
          break
        else
          vb.views["mono button"].color = COLOR_GREY
          break
        end  
      end
    end
    

    --If filter is found on reference track get its state/ preset and set the button colors accordingly 
    ---------------------------------------------------------------------------------------------------
    if (ref_track ~= nil) and (ref_filter ~= nil) then
       
      --only high and low visible
      if vb.views["mid checkbox"].value == false then
         
        --(ref_filter.parameters[2].value)
         
        if ref_filter.parameters[2].value == low_filter_cutoff_mid_disabled_num and (ref_filter.is_active) then
          vb.views["low button"].color = COLOR_ORANGE
          vb.views["high button"].color = COLOR_GREY
        elseif ref_filter.parameters[2].value == high_filter_cutoff_mid_disabled_num and (ref_filter.is_active) then
          vb.views["low button"].color = COLOR_GREY
          vb.views["high button"].color = COLOR_ORANGE
        else
          vb.views["low button"].color = COLOR_GREY
          vb.views["high button"].color = COLOR_GREY
        end
       
      
      else--all three buttons are active
        -- print(ref_filter.parameters[2].value)
         
        -- print("mid filter cutoff: "..tostring(mid_filter_cutoff))
        if (ref_filter.parameters[2].value == low_filter_cutoff_all_enabled_num) and (ref_filter.is_active) then--== LOW_CUTOFF then  --0.125  is low filter cutoff value
          vb.views["low button"].color = COLOR_ORANGE
        else
          vb.views["low button"].color = COLOR_GREY
        end
       -- if (ref_filter.parameters[2].value > 0.3) and (ref_filter.parameters[2].value < 0.4) and (ref_filter.is_active) then-- == MID_CUTOFF then
       if ref_filter.parameters[2].value == (mid_filter_cutoff_num) and (ref_filter.is_active) then
          vb.views["mid button"].color = COLOR_ORANGE
        else
         vb.views["mid button"].color = COLOR_GREY
        end
        if ref_filter.parameters[2].value == (high_filter_cutoff_all_enabled_num) and (ref_filter.is_active) then-- == HIGH_CUTOFF then
          vb.views["high button"].color = COLOR_ORANGE
        else
          vb.views["high button"].color = COLOR_GREY
        end
      end
     end
    --update volume slider
    vb.views["volume slider"].value = ref_track.postfx_volume.value
  end
  
  --3)CASE DIALOG CLOSED
  ------------------------------------
  --remove timer when dialog is closed
  ------------------------------------
  if my_dialog and (my_dialog.visible == false) then
    --remove timer
    if renoise.tool():has_timer(timer) then
     renoise.tool():remove_timer(timer)
    end
    
    -----TIDY UP DEVICES THAT THE TOOL ADDED (FILTER AND MONO)---
    
    --get master track object
    local master_index = #song.tracks - song.send_track_count
    local master_track = song:track(master_index)
    
    --Delete Reference filter on master track
    for devs = #master_track.devices,2 ,-1 do
      if master_track.devices[devs].display_name == "Reference Filter" then
        master_track:delete_device_at(devs)
        break
      end
    end
    --Delete Mono Mix device on master
    for devs = 2, #master_track.devices do
      if (master_track:device(devs).display_name == "Mono Mix") then 
       master_track:delete_device_at(devs)
       break
      end
    end
    ---------------------------------------------
    --Delete Reference filter on ref treack track
    if ref_track_idx ~= nil then
      local ref_track = song:track(ref_track_idx)
      for devs = #ref_track.devices,2 ,-1 do
        if ref_track.devices[devs].display_name == "Reference Filter" then
          ref_track:delete_device_at(devs)
          break
        end
      end
    end
    --Delete Mono Mix device on ref
    if ref_track_idx ~= nil then
      local ref_track = song:track(ref_track_idx)
      for devs = 2, #ref_track.devices do
        if (ref_track:device(devs).display_name == "Mono Mix") then 
         ref_track:delete_device_at(devs)
         break
        end
      end
    end
    ---------------------------------------------
  end --end of dialog closed
  
  timer_running = false
end--end of timer
-------------------------------------------------
-------------------------------------------------
